{
	xEdit Auto-patch script to enable some future functionality for Gear Grinder 2.0
	Applies components from OMOD COBJs to their loose MISC items.
	Usage:
		Save to your xEdit\Edit Scripts folder
		Load xEdit with your full load order
		Right click anywhere in the left pane -> Apply Script...
		Find 'GearGrinderMiscComponents' in the dropdown
		Click OK
		Wait a bit for the script to finish (takes about 15s on my machine)
		Save changes to GearGrinder_AutoPatch.esp
		Close xEdit
}
unit GearGrinderMiscComponents;

const
	patchName = 'GearGrinder2_AutoPatch.esp';
var
	total: Cardinal;
	patch: IInterface;
	tempMasterList: TStringList;
	addedMasters : THashedStringList;
	UnscrappableObject, ObjectTypeLooseMod: IInterface;

function HasKeyword(e: IInterface; kywd: IInterface): boolean;
var
	keywords: IInterface;
	i: Cardinal;
begin
	keywords := ElementBySignature(e, 'KWDA');
	for i := 0 to ElementCount(keywords) - 1 do begin
		if Equals(LinksTo(ElementByIndex(keywords, i)), kywd) then begin
			Result := True;
			Exit;
		end;
	end;
	Result := False;
end;

procedure AddKeyword(e: IInterface; kywd: IInterface);
var
	keywords, element: IInterface;
	i: Cardinal;
begin
	if not ElementExists(e, 'KWDA') then
		keywords := Add(e, 'KWDA', False)
	else
		keywords := ElementBySignature(e, 'KWDA');
	
	element := ElementAssign(keywords, HighInteger, nil, False);
	SetNativeValue(element, GetLoadOrderFormID(kywd));
end;

function ProcessRecord(e: IInterface): integer;
var
	master: string;
	omod, misc, miscOverride, srcCmpoArr, destCmpoArr, cmpoStruct, cmpo: IInterface;
	i: Cardinal;
begin
	if not IsWinningOverride(e) then
		Exit;
	
	// get passed-in COBJ's OMOD object
	omod := ElementBySignature(e, 'CNAM');
	if Equals(omod, Nil) then Exit;
	omod := WinningOverride(LinksTo(omod));
	if Signature(omod) <> 'OMOD' then Exit;

	// get OMOD object's MISC item
	misc := ElementBySignature(omod, 'LNAM');
	if Equals(misc, Nil) then Exit;
	misc := WinningOverride(LinksTo(misc));
	if Signature(misc) <> 'MISC' then Exit;
	
	// skip record if predefined CMPO data already exists on the MISC item
	destCmpoArr := ElementBySignature(misc, 'CVPA');
	if destCmpoArr <> Nil then Exit; 
	
	// add masters required for by COBJ/OMOD/MISC to patch
	ReportRequiredMasters(e, tempMasterList, True, True);
	ReportRequiredMasters(omod, tempMasterList, True, True);
	ReportRequiredMasters(misc, tempMasterList, True, True);
	
	for i := 0 to tempMasterList.Count - 1 do begin
		master := tempMasterList[i];
		if addedMasters.IndexOf(master) < 0 then begin
			// AddMessage('need to add ' + tempMasterList[i]);
			AddMasterIfMissing(patch, master);
			addedMasters.Add(master);
		end;
	end;
	tempMasterList.Clear;
	
	// copy MISC item as override into patch
	miscOverride := wbCopyElementToFile(misc, patch, False, True);
	
	// check that MISC item has ObjectTypeLooseMod keyword
	if not HasKeyword(miscOverride, ObjectTypeLooseMod) then begin
		AddMessage('Warning #1: MISC loose mod item ' + Name(miscOverride) + ' missing ObjectTypeLooseMod keyword--this error will be patched');
		AddKeyword(miscOverride, ObjectTypeLooseMod);
	end;
	
	AddKeyword(miscOverride, UnscrappableObject);
	
	// assign+sanity check COBJ's CMPO data and copy it to MISC item
	srcCmpoArr := ElementBySignature(e, 'FVPA');
	Add(miscOverride, 'CVPA', False);
	destCmpoArr := ElementBySignature(miscOverride, 'CVPA');
	for i := 0 to ElementCount(srcCmpoArr) - 1 do begin
		cmpoStruct := ElementByIndex(srcCmpoArr, i);
		cmpo := LinksTo(ElementByPath(cmpoStruct, 'Component'));
		if Signature(cmpo) = 'CMPO' then
			ElementAssign(destCmpoArr, HighInteger, cmpoStruct, False)
		else
			AddMessage('Warning #2: COBJ ' + Name(e) + ' recipe entry ' + IntToStr(i) + ':' + Name(cmpo) + ' not a CMPO--skipping');
	end;
	
	inc(total);
	Exit;
end;

function ProcessGroup(e: IInterface): integer;
var
	i: Integer;
begin
	for i := 0 to ElementCount(e) - 1 do begin
		ProcessRecord(ElementByIndex(e, i));
	end;
end;

function Initialize: integer;
var
	iFileIndex: Integer;
	iFile: IInterface;
begin
	tempMasterList := TStringList.Create;
	addedMasters := THashedStringList.Create;
	
	UnscrappableObject := RecordByFormID(FileByIndex(0), $1CC46A, True); // UnscrappableObject [KYWD:001CC46A]
	ObjectTypeLooseMod := RecordByFormID(FileByIndex(0), $135C17, True); // ObjectTypeLooseMod [KYWD:00135C17]
	
	for iFileIndex := 0 to FileCount - 1 do begin
		iFile := FileByIndex(iFileIndex);
		if (GetFileName(iFile) = patchName) then begin
			patch := iFile;
			break
		end;
	end;
	
	if not Assigned(patch) then begin
		patch := AddNewFileName(patchName, True);
	end;
	
	RemoveNode(GroupBySignature(patch, 'MISC'));
	
	for iFileIndex := 0 to FileCount - 1 do begin
		iFile := FileByIndex(iFileIndex);
		if GetFileName(iFile) <> patchName then
			ProcessGroup(GroupBySignature(iFile, 'COBJ'));
	end;
	Result := 0;
end;

function Finalize: integer;
begin
	tempMasterList.Free;
	addedMasters.Free;
	
	CleanMasters(patch);
	
	AddMessage('Done. Patched ' + IntToStr(total) + ' records.');
	Result := 0;
end;

end.
